#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# 
# Copyright © 2007-2009 Frederik Elwert

import os
import sys
import glob
import thread
import logging
import textwrap

# Check if this file is run directly, not as a module.
# Insert the parent directory to use source files instead of installed ones.
if __name__ == '__main__':
    os.chdir(os.path.realpath(os.path.dirname(sys.argv[0])))
    os.chdir('../')
    sys.path.insert(0, os.getcwd())

import pygtk
pygtk.require('2.0')
import gtk
import gtk.glade
import gobject
import pynotify
import dbus
import locale
import gettext
import configobj
import validate
from xdg import BaseDirectory, DesktopEntry

from Genesis.config import *
from Genesis import syncevolution
from Genesis import settings
from Genesis import __version__, __copyright__
## Setup ##
# l10n
locale.setlocale(locale.LC_ALL, '')
for module in (gettext, gtk.glade):
    module.bindtextdomain(APP_NAME, LOCALE_DIR)
    module.textdomain(APP_NAME)
_ = gettext.gettext

# Logging
logger = logging.getLogger('genesis')

# notify
pynotify.init('Genesis')

# dbus (NetworkManager)
NM_STATE_UNKNOWN = 0
NM_STATE_ASLEEP = 1
NM_STATE_CONNECTING = 2
NM_STATE_CONNECTED = 3
NM_STATE_DISCONNECTED = 4

# display result options
SHOW_ALL = 3
SHOW_NONEMPTY = 2
SHOW_ERROR = 1
SHOW_NONE = 0

bus = dbus.SystemBus()
try:
    nm = bus.get_object('org.freedesktop.NetworkManager',
            '/org/freedesktop/NetworkManager')
    # Test if the state can be requested
    nm.state()
except dbus.exceptions.DBusException:
    nm = None

def nonzero(sequence):
    """
    Look if any value in a sequence is nonzero.
    
    """
    result = False
    for item in sequence:
        if item != 0:
            result = True
            break
    return result

def format_result(result):
    """
    Format the result of a sync action for display.
    
    :Parameters:
        - `result`: a dict as returned by SyncEvolution
        
    """
    messages = [
            _('new on client'),
            _('modified on client'),
            _('deleted on client'),
            _('new on server'),
            _('modified on server'),
            _('deleted on server')
            ]
    output = ''
    for key in result:
        if True in [nonzero(item) for item in result[key]]:
            output += '<b>%s:</b>\n' % key
            for num, item in enumerate(result[key]):
                if item[0] > 0:
                    # wrap text to respect restricted room in notification
                    lines = textwrap.wrap('%s %s' % (item[0], messages[num]),
                                          20)
                    output += '\n'.join(['\t' + line for line in lines])
                    output += '\n'
    if not output:
        output = _('No changes necessary.')
    return output


class AnimIcon(object):
    """
    An animatable icon.
    
    """
    
    def __init__(self, image, rate):
        """
        :Parameters:
            - `image`: A widget that implements ``set_from_pixbuf``.
            - `rate`: frames per seconds as int.
            
        """
        self.image = image
        self.rate = rate
        self.pixbufs = []
        self.position = 0
        self._static_image = None
        self._animate = False
        
    def get_animate(self):
        return self._animate
        
    def set_animate(self, state):
        logger.debug('Animate icon: %s' % state)
        self._animate = state
        if self._animate == True:
            delay = 1000 // self.rate
            gobject.timeout_add(delay, self.update_image,
                    priority=gobject.PRIORITY_HIGH)
        else:
            self.update_image()
            
    animate = property(get_animate, set_animate)
            
    def update_image(self):
        if self._animate:
            self.advance()
            self.image.set_from_pixbuf(self.get_pixbuf())
        else:
            self.image.set_from_pixbuf(self.static_image)
        return self.animate
        
    def get_static_image(self):
        if self._static_image:
            return self._static_image
        else:
            return self.pixbufs[0]
            
    def set_static_image(self, pixbuf=None):
        """
        Set the static image.
        
        :Parameters:
            - `pixbuf`: a gtk.Pixbuf.
        
        """
        self._static_image = pixbuf

    static_image = property(get_static_image, set_static_image)

    def advance(self):
        self.position += 1
        if self.position >= len(self.pixbufs):
            self.position = 0
            
    def get_pixbuf(self):
        return self.pixbufs[self.position]
        
    def add_from_file(self, filename):
        pixbuf = gtk.gdk.pixbuf_new_from_file(filename)
        self.pixbufs.append(pixbuf)

    
class GenesisConfig(object):
    """
    Class to store and edit Genesis' general preferences.
    
    """
    _bools = {
        'yes': True, 'no': False,
        'on': True, 'off': False,
        '1': True, '0': False,
        'true': True, 'false': False,
        }

    def __init__(self):
        config_file = os.path.join(CONFIG_DIR, APP_NAME + '.ini')
        spec_file = os.path.join(DATA_DIR, APP_NAME + '.spec.ini')
        validator = validate.Validator()
        self.config = configobj.ConfigObj(config_file, configspec=spec_file)
        self.config.validate(validator, copy=True)
        
    def __setitem__(self, name, value):
        if name == 'autostart':
            self.set_autostart(value)
        else:
            self.config[name] = value
            self.config.write()
            
    def __getitem__(self, name):
        if name == 'autostart':
            return self.get_autostart()
        else:
            return self.config[name]

    def _get_autostart_file(self):
        """
        Get the path of the autostart desktop file.
        
        """
        desktop_path = None
        for path in BaseDirectory.load_config_paths(AUTOSTART_DIR):
            desktop = os.path.join(path, DESKTOP_FILE)
            if os.path.isfile(desktop):
                desktop_path = desktop
                break
        return desktop_path
        
    def _as_bool(self, value):
        """
        Transform a config value into a boolean.
        
        The following values will return True: yes, true, on, 1.
        All other values will return False.
        
        This method is inspired from configobj.
        
        :Parameters:
            - `value`: The value to transform.
            
        """
        if value is True:
            return True
        elif value is False:
            return False
        else:
            try:
                return self._bools[value.lower()]
            except KeyError:
                return False
            
    def set_autostart(self, value):
        """
        Enable or disable the autostart of the application.
        
        :Parameters:
            - `value`: A boolean.
        
        """
        desktop_path = self._get_autostart_file()
        if not desktop_path:
            desktop_path = os.path.join(DESKTOP_DIR, DESKTOP_FILE)
        config = DesktopEntry.DesktopEntry(desktop_path)
        config.filename = os.path.join(BaseDirectory.xdg_config_home,
                AUTOSTART_DIR, DESKTOP_FILE)
        config.set('X-GNOME-Autostart-enabled', str(value).lower())
        config.write()
        
    def get_autostart(self):
        """
        Returns if autostart is enabled.
        
        :Return:
            - A boolean.
        
        """
        desktop_path = self._get_autostart_file()
        if not desktop_path:
            return False
        config = DesktopEntry.DesktopEntry(desktop_path)
        active = config.get('X-GNOME-Autostart-enabled')
        active = self._as_bool(active)
        return active
            

class PreferencesDialog(object):
    """
    The dialog to edit Genesis' general preferences.
    
    """

    def __init__(self, config, widgets):
        """
        :Parameters:
            - `config`: a GenesisConfig object to store the preferences.
            - `widgets`: a list of widget names corresponding to option names.
            
        """
        self.config = config
        
        # GTK/Glade setup
        self.wTree = gtk.glade.XML(
                os.path.join(DATA_DIR, 'settings.glade'), 'preferences_dialog')
        signals = {
            'on_preference_change': self.save_preference,
        }
        self.wTree.signal_autoconnect(signals)
        self.dialog = self.wTree.get_widget('preferences_dialog')
        
        self.widgets = {}
        for widget in widgets:
            self.widgets[widget] = self.wTree.get_widget(widget)

    def save_preference(self, widget):
        """
        Pass the preference value to the config object.
        
        """
        widget_type, name = widget.name.split('_', 1)
        if hasattr(widget, 'get_value'):
            value = widget.get_value()
        elif hasattr(widget, 'get_active'):
            value = widget.get_active()
        self.config[name] = value
        
    def run(self):
        # Load configuration into dialog
        for widget_name in self.widgets:
            widget = self.widgets[widget_name]
            widget_type, name = widget_name.split('_', 1)
            # if config key exists, load into dialog
            try:
                value = self.config[name]
                if hasattr(widget, 'set_value'):
                    widget.set_value(value)
                elif hasattr(widget, 'set_active'):
                    widget.set_active(value)
            # otherwise, write dialog default into config
            except KeyError:
                self.save_preference(widget)
        ret = self.dialog.run()
        self.dialog.destroy()
        
    
class Genesis(object):
    """
    The GUI class for Genesis.
    
    """
    def __init__(self, status_icon):
        # Configuration
        self.config = GenesisConfig()
        try:
            self.server = self.config['default_server']
        except KeyError:
            self.server = None
        self.log = None
        self.syncevolution = syncevolution.SyncEvolution()
        self.tooltip = _('Synchronize with "%s"')
        self.tooltip_empty = _('Configure SyncEvolution')
        self.tooltip_sync = _('Synchronizing with "%s"')
        
        # Init the status icon
        self.status_icon = status_icon
        self.status_icon.set_visible(True)
        
        # Init the animated icon
        self.anim_icon = AnimIcon(self.status_icon, 10)
        icon_files = glob.glob(os.path.join(IMAGE_DIR, 'genesis-sync??.png'))
        icon_files.sort()
        for icon in icon_files:
            self.anim_icon.add_from_file(icon)
        self.anim_icon.set_static_image(
                gtk.gdk.pixbuf_new_from_file(
                os.path.join(IMAGE_DIR, 'genesis.png'))
                )
        self.anim_icon.update_image()
        
        # Small hack to make sure the icon is already there when we
        # show startup messages
        gobject.timeout_add(100, self.build_menu)
        
    def build_menu(self):
        self.menu = gtk.Menu()
        # Sync
        sync_action = gtk.Action('sync', _('Synchronize'),
                _('Start synchronization with server'),
                gtk.STOCK_CONVERT)
        self.sync_menu_item = sync_action.create_menu_item()
        self.sync_menu_item.connect('activate', self.sync)
        self.menu.append(self.sync_menu_item)
        # Special syncs
        self.syncs_menu = gtk.Menu()
        sync_modes = ['two-way', 'slow',
                      'refresh-from-client', 'refresh-from-server',
                      'one-way-from-client', 'one-way-from-server']
        for mode in sync_modes:
            menuItem = gtk.MenuItem(mode)
            menuItem.connect('activate', self.sync, mode)
            self.syncs_menu.append(menuItem)
        menuItem = gtk.MenuItem(_('Special Synchronization'))
        menuItem.set_submenu(self.syncs_menu)
        self.menu.append(menuItem)
        # Select servers
        self.servers_menu = gtk.Menu()
        self.choose_server = gtk.MenuItem(_('Select Server'))
        self.choose_server.set_submenu(self.servers_menu)
        self.menu.append(self.choose_server)
        # Add server
        add_action = gtk.Action('add_server', _('Add Server...'),
                _('Add a server to the configuration'),
                gtk.STOCK_ADD)
        menuItem = add_action.create_menu_item()
        menuItem.connect('activate', self.add_server)
        self.menu.append(menuItem)
        self.build_server_menu()
        # Separator
        menuItem = gtk.SeparatorMenuItem()
        self.menu.append(menuItem)
        # Preferences
        menuItem = gtk.ImageMenuItem(gtk.STOCK_PREFERENCES)
        menuItem.connect('activate', self.show_preferences)
        self.menu.append(menuItem)
        # Separator
        menuItem = gtk.SeparatorMenuItem()
        self.menu.append(menuItem)
        # About
        menuItem = gtk.ImageMenuItem(gtk.STOCK_ABOUT)
        menuItem.connect('activate', self.show_about)
        self.menu.append(menuItem)
        # Quit
        menuItem = gtk.ImageMenuItem(gtk.STOCK_QUIT)
        menuItem.connect('activate', self.quit)
        self.menu.append(menuItem)
        
        self.status_icon.connect('popup-menu', self.show_menu,
                self.menu)
        self.status_icon.connect('activate', self.sync)

        # Startup options
        if self.config['sync_on_startup']:
            self.sync(None)
        self.init_autosync()

        return False
        
    def quit(self, widget):
        """
        Exit the application.
        
        """
        gtk.main_quit()
        
    def build_server_menu(self):
        """
        Look for available servers and build a menu from them.
        
        """
        # remove current menu items
        for child in self.servers_menu.get_children():
            self.servers_menu.remove(child)
        # Look for available server configs
        self.servers = self.syncevolution.list_servers().keys()
        # Are there any servers?
        if not self.servers:
            # inticate that syncing is not possible
            self.sync_menu_item.set_sensitive(False)
            self.choose_server.set_sensitive(False)
            self.status_icon.set_tooltip(self.tooltip_empty)
            self.notify(_('Configure SyncEvolution'), 
                    _('''SyncEvolution is not yet configured.
To set up synchronization, please click the icon.'''))
            return
        self.servers.sort()
        if not self.server:
            self.server = self.servers[0]
        self.sync_menu_item.set_sensitive(True)
        self.choose_server.set_sensitive(True)
        self.status_icon.set_tooltip(self.tooltip % self.server)
        # Add server items to the menu
        self.servers_items = []
        for server in self.servers:
            group = None
            if self.servers_items:
                group = self.servers_items[0]
            menuItem = gtk.RadioMenuItem(group, server, False)
            if server == self.server:
                menuItem.set_active(True)
            self.servers_items.append(menuItem)
            self.servers_menu.append(menuItem)
            menuItem.connect('toggled', self.select_server)
            
    def add_server(self, widget):
        """
        Call the server assistant to add a new server.
        
        """
        server_assistant = settings.ServerAssistant(caller=self)
        
    def show_preferences(self, widget):
        widgets = ['checkbutton_autostart', 'checkbutton_sync_on_startup',
                'checkbutton_autosync', 'spinbutton_sync_interval']
        preferences = PreferencesDialog(self.config, widgets)
        preferences.run()
        self.init_autosync()
        
    def show_menu(self, widget, button, time, data = None):
        if button == 3:
            if data:
                data.show_all()
                data.popup(None, None, None, 3, time)
                
    def show_about(self, widget):
        wTree = gtk.glade.XML(os.path.join(DATA_DIR, 'genesis.glade'), 'aboutdialog')
        about = wTree.get_widget('aboutdialog')
        logo = gtk.gdk.pixbuf_new_from_file(
                os.path.join(IMAGE_DIR, 'genesis.svg'))
        about.set_logo(logo)
        about.set_version(__version__)
        about.set_copyright(__copyright__)
        about.run()
        about.destroy()
        
    def select_server(self, widget):
        """
        Set the server the user chooses in the menu.
        
        """
        if widget.get_active():
            index = self.servers_items.index(widget)
            self.server = self.servers[index]
            # save the selected server in the config file
            self.config['default_server'] = self.server
            self.status_icon.set_tooltip(self.tooltip % self.server)
            
    def notify(self, title, message, type='message'):
        """
        Show a notification window.
        
        :Parameters:
            - `title`: The title of the notification.
            - `message`: The message body.
            - `type`: 'message' or 'error'.
            
        """
        logger.debug('Notification:\n    %s\n    %s\n\n    %s' % (
                  title, '-' * len(title), message))
        icon = uri = 'file://' + os.path.join(os.path.abspath(IMAGE_DIR), 'genesis.svg')
        if type == 'error':
            icon = 'dialog_error'
        self.notification = pynotify.Notification(title, message, icon)
        self.notification.attach_to_status_icon(self.status_icon)
        if self.log:
            self.notification.add_action('show_log', _('Show Log'),
                    self.show_log)
            print self.log
        self.notification.show()
        
    def show_log(self, notification, action):
        """
        Show the error log if sync failed.
        
        """
        assert action == 'show_log'
        wTree = gtk.glade.XML(os.path.join(DATA_DIR, 'genesis.glade'), 'log_win')
        self.log_win = wTree.get_widget('log_win')
        textview = wTree.get_widget('textview')
        textbuffer = textview.get_buffer()
        textbuffer.set_text(self.log)
        button = wTree.get_widget('close_button')
        button.connect('clicked', self.hide_log)
        self.log_win.show_all()
        
    def hide_log(self, widget):
        self.log_win.destroy()
        
    def init_autosync(self):
        # remove previous timeouts
        try:
            gobject.source_remove(self.autosync_id)
        except AttributeError:
            pass
        # add timeout
        if self.config['autosync']:
            interval = int(self.config['sync_interval'] * 60)
            self.autosync_id = gobject.timeout_add_seconds(
                    interval,
                    self.sync, None)
        
    def sync(self, widget, *args):
        """
        Perform the synchronization.
        
        Additional arguments are passed to SyncEvolution.sync()
        
        """
        logger.info('Starting sync with %s' % self.server)
        self.log = None
        # Don't show empty results if sync is started automatically
        if widget is None:
            show = SHOW_NONEMPTY
        else:
            show = SHOW_ALL
        if not self.server:
            self.add_server(widget)
            return
        self.anim_icon.set_animate(True)
        self.status_icon.set_tooltip(self.tooltip_sync % self.server)
        thread.start_new_thread(
                    self.sync_thread, args, {'show': show})
        return True
        
    def sync_thread(self, *args, **kwargs):
        """
        Call syncevolution and report the result.
        
        :Parameters:
            - `show`: Whether the result of changes should be displayed.
              Currently supported values:
              
              ``SHOW_ALL``
                  All results are displayed.
                  
              ``SHOW_NONEMPTY``
                  Results are only displayed, if changes have been performed.
        
        """
        if 'show' in kwargs:
            show = kwargs['show']
        try:
            # Assume working network connection if networkmanager is not
            # available.
            if nm is None:
                network_available = True
            # Otherwise, ask networkmanager for network state.
            else:
                network_available = nm.state() == NM_STATE_CONNECTED
            # Only try syncing if network is available
            if network_available:
                result = self.syncevolution.sync(self.server, *args)
                logger.debug('Sync result: %s' % result)
            else:
                raise syncevolution.SyncError(_('Network not available.'))
        except syncevolution.SyncError, e:
            self.log = e.log
            gtk.gdk.threads_enter()
            self.notify(_('Synchronization Failed'), e.message, 'error')
            logger.error('Sync failed: %s' % e.message)
            gtk.gdk.threads_leave()
        else:
            # See if a sync has actually been performed
            sync_performed = False
            for key in result:
                if True in [nonzero(item) for item in result[key]]:
                    sync_performed = True
                    break
            if show == SHOW_ALL or (show == SHOW_NONEMPTY and sync_performed):
                gtk.gdk.threads_enter()
                self.notify(_('Synchronization Results'),
                        format_result(result))
                gtk.gdk.threads_leave()
        gtk.gdk.threads_enter()
        self.anim_icon.set_animate(False)
        self.status_icon.set_tooltip(self.tooltip % self.server)
        gtk.gdk.threads_leave()
        
        
def main():
    si = gtk.StatusIcon()
    g = Genesis(si)
    gtk.gdk.threads_init()
    gtk.main()

if __name__ == '__main__':
    main()

