#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# 
# Copyright © 2007-2009 Frederik Elwert

from StringIO import StringIO
import re
import os
import subprocess
import locale
import gettext
import ConfigParser
from cStringIO import StringIO
import logging

from Genesis.config import *

## Setup ##
# l10n
locale.setlocale(locale.LC_ALL, '')
gettext.bindtextdomain(APP_NAME, LOCALE_DIR)
gettext.textdomain(APP_NAME)
_ = gettext.gettext

# Logging
logger = logging.getLogger('syncevolution')
logger.debug('Importing syncevolution')

class SyncError(Exception):
    """
    Synchronisation error.
    
    """
    def __init__(self, message, log=None):
        Exception.__init__(self)
        self.message = message
        self.log = log

class SyncEvolution(object):
    """
    A binding for syncevolution.
    
    This class provides a basic binding for syncevolution. In fact, it
    calls the syncevolution binary directly, but makes it options
    available and pareses it's output.
    
    """

    @classmethod
    def _parse_result(cls, fo):
        """
        Parses the result of a synchronisation.
        
        :Parameters:
            - `fo`: a file object or a string.
            
        :Return:
            - a dictionary that reports results for each source.
            - the transaction log as a string.
        
        """
        # Create file-like object if string is passed.
        if not hasattr(fo, 'readline'):
            fo = StringIO(fo)
        # Search for beginning of log lines (starting with timestamp)
        logline = re.compile('\[[A-Z]+\]')
        # write log messages to file or StringIO
        log = StringIO()
        islog = False
        while True:
            line = fo.readline()
            if line == '':
                break
            if logline.match(line) is not None:
                islog = True
            if line.strip() == '' and islog is True:
                break
            if islog:
                log.write(line)
        log_value = log.getvalue()
        log.close()
        # look for update table and parse it
        # table rows start with something like
        # +------+-----+---------+
        data_row = re.compile('^\|[a-z ]+[\|0-9 \/]+\|$')
        changes = {}
        while True:
            line = fo.readline()
            if line == '':
                break
            if data_row.match(line) is not None:
                # get table cells
                data = [item.strip()
                        for item in line.split('|')
                        if item.strip() is not '']
                # get key
                key = data.pop(0)
                # split successful/total
                data = [item.split('/') for item in data if item.find('/') > -1]
                # Turn from strings into ints
                # item[0] changed its meaning from "successful" to "rejected"
                # in syncevolution 0.9, so using item[1]-item[0] to get the old
                # meaning.
                data = [(int(item[1])-int(item[0]), int(item[1]))
                        for item in data]
                changes[key] = data
        return changes, log_value
        
    @classmethod
    def _parse_to_dict(cls, lines):
        """
        Parse some of syncevolution's standard outputs to a dict.
        
        Some options of syncevolution output values as "key = value" pairs.
        Parse these and return a ``{key: value}`` dict.
        
        :Parameters:
            - `lines`: An interable holding the output lines.
            
        :Returns:
            - A dict.
        
        """
        regex = re.compile(r'(.+) = (.+)')
        result_dict = {}
        for line in lines:
            match = regex.search(line)
            if match is not None:
                 result_dict[match.group(1).strip()] = match.group(2).strip()
        return result_dict
        
    @classmethod
    def _get_version(cls):
        """
        Print the version of syncevolution and return a version tuple.
        
        :Returns:
            - a version tuple.
            
        """
        version = subprocess.Popen(
                ['syncevolution', '--version'],
                stdout=subprocess.PIPE
                )
        line = version.stdout.readline()
        match = re.search(r'((\d+\.?)+)', line)
        version_tuple = tuple(match.group(1).split('.'))
        version.wait()
        version.stdout.close()
        return version_tuple
        
    version = property(_get_version)
        
    @classmethod
    def sync(cls, server, mode=None):
        """
        Perform the sync.
        
        :Parameters:
            - `server`: The server that should be synced.
            - `mode`: The sync mode for this run. If ``None``, standard sync
              mode is used.
        
        """
        command = ['syncevolution']
        if mode is not None:
            logger.debug('Sync mode: %s' % mode)
            command.extend(['--sync', mode])
        command.append(server)
        sync = subprocess.Popen(
                command,
                stdout = subprocess.PIPE
                )
        stdout, stderr = sync.communicate()
        changes, log = cls._parse_result(stdout)
        result = sync.poll()
        if result != 0:
            raise SyncError(_('An error occured during the synchronisation process.'), log)
        return changes
        
    @classmethod
    def list_servers(cls):
        """
        List all configured servers.
        
        Lists all servers configured for use with syncevolution. This
        corresponds to syncevolution's ``--print-servers`` option.
        
        :Returns:
            - A dict of server name to config directory mapping.
        
        """
        print_servers = subprocess.Popen(
                ['syncevolution', '--print-servers'],
                stdout = subprocess.PIPE
                )
        stdout, stderr = print_servers.communicate()
        server_dict = cls._parse_to_dict(stdout.splitlines())
        return server_dict
    
    @classmethod
    def list_templates(cls):
        """
        List all available config templates.
        
        :Returns:
            - A dict of template name and server url.
        
        """
        templates = subprocess.Popen(
                ['syncevolution', '--template', '?'],
                stdout = subprocess.PIPE
                )
        stdout, stderr = templates.communicate()
        template_dict = cls._parse_to_dict(stdout.splitlines())
        return template_dict
    

class ServerConfig(object):
    """
    This class holds the configuration for a server.
    
    """

    def __init__(self, name, template=None):
        """
        Load a configuration or create a new one.
        
        :Parameters:
            - `name`: The name of the server. If the server doesn't
              already exist, a new one is created using `template`.
            - `template`: The name of a configuration template.
        
        """
        self.name = name
        self.template = template
        self.server_config = {}
        self.source_config = {}
        
    def load_template(self, template=None):
        """
        Load the config from a template.
        
        :Parameters:
            - `template`: The name of the template.
        
        """
        # If no template is given, use this configs template
        if template is None:
        # If there is no template for this config, stop here
            template = self.template
        if template is None:
            return
            
        command = ['syncevolution', '--print-config', '--quiet',
                   '--template', template]
        process = subprocess.Popen(command, stdout=subprocess.PIPE)
        output = process.stdout.readlines()
        process.wait()
        # syncevolution outputs the server config without section header
        # So we insert a default section header at the beginning
        output.insert(0, '[server]\n')
        config_fp = StringIO(''.join(output))
        config = ConfigParser.RawConfigParser()
        config.optionxform = str
        config.readfp(config_fp)
        self.server_config = dict(config.items('server'))
        for section in config.sections():
            if section != 'server':
                self.source_config[section] = dict(config.items(section))

    def add_source(self, source):
        """
        Add a source to the configuration.
        
        :Parameters:
            - `source`: The source name.
            
        """
        self.source_config[source] = {}
        
    def save(self):
        """
        Save the configuration.
        
        """
        # basic command
        command = ['syncevolution', '--configure']
        # use a template if given
        if self.template is not None:
            command.extend(['--template', self.template])
        # append '--sync-property option=value' for all server options
        for option in self.server_config:
            command.extend(['--sync-property', '%s=%s' % (
                    option, self.server_config[option])])
        # append server name
        command.append(self.name)
        # call the command for server setup
        logger.info('Set up a new server: %s' % self.name)
        logger.debug('Command: %s' % command)
        retcode = subprocess.call(command)
        if retcode != 0:
            logger.error('Setting up %s failed with code %s!'
                         % (self.name, retcode))
            return
        # Built commands for source setup quite the same way as for server setup
        for source in self.source_config:
            command = ['syncevolution', '--configure']
            for option in self.source_config[source]:
                command.extend(['--source-property', '%s=%s' % (
                        option, self.source_config[source][option])])
            command.extend([self.name, source])
            logger.info('Configuring source %s' % source)
            logger.debug('Command: %s' % command)
            retcode = subprocess.call(command)
            if retcode != 0:
                logger.error('Configuring %s failed with code %s!'
                             % (self.name, retcode))
                return
                            

if __name__ == '__main__':
    result_file = open('sync.txt')
    sync = SyncEvolution()
    print sync._parse_result(result_file)
    result_file.close()

